@@ -1,8 +1,8 @@
 from .registered_task import RegisteredTask, MockTask, PrintTask
-from .taskqueue import TaskQueue, MockTaskQueue, LocalTaskQueue, upload
+from .taskqueue import TaskQueue, GreenTaskQueue, MockTaskQueue, LocalTaskQueue, upload
 from .secrets import (
   QUEUE_NAME, TEST_QUEUE_NAME, QUEUE_TYPE, 
   PROJECT_NAME, AWS_DEFAULT_REGION
 )
 
-__version__ = '0.12.3'
\ No newline at end of file
+__version__ = '0.13.0'
\ No newline at end of file
