@@ -12,14 +12,17 @@ import time
 import traceback
 
 import concurrent.futures 
+import gevent.pool
 import multiprocessing as mp
 import numpy as np
 from tqdm import tqdm
 
 from cloudvolume.threaded_queue import ThreadedQueue
+from cloudvolume.lib import yellow, scatter
 
 from .aws_queue_api import AWSTaskQueueAPI
 from .registered_task import RegisteredTask, deserialize
+from .scheduler import schedule_green_jobs, schedule_threaded_jobs
 from .secrets import (
   PROJECT_NAME, QUEUE_NAME, QUEUE_TYPE,
   AWS_DEFAULT_REGION
@@ -30,31 +33,17 @@ def totask(task):
   taskobj._id = task['id']
   return taskobj
 
+class QueueEmpty(LookupError):
+  def __init__(self):
+    super(LookupError, self).__init__('Queue Empty')
+
 LEASE_SECONDS = 300
 
-class TaskQueue(ThreadedQueue):
-  """
-  The standard usage is that a client calls lease to get the next available task,
-  performs that task, and then calls task.delete on that task before the lease expires.
-  If the client cannot finish the task before the lease expires,
-  and has a reasonable chance of completing the task,
-  it should call task.update before the lease expires.
-
-  If the client completes the task after the lease has expired,
-  it still needs to delete the task. 
-
-  Tasks should be designed to be idempotent to avoid errors 
-  if multiple clients complete the same task.
-  """
-  class QueueEmpty(LookupError):
-    def __init__(self):
-      super(LookupError, self).__init__('Queue Empty')
-
+class SuperTaskQueue(object):
   def __init__(
     self, queue_name=QUEUE_NAME, queue_server=QUEUE_TYPE, 
     region=None, qurl=None, n_threads=40, project=PROJECT_NAME
   ):
-
     self._project = project
     self._region = region
     self._queue_name = queue_name
@@ -62,16 +51,6 @@ class TaskQueue(ThreadedQueue):
     self._qurl = qurl
     self._api = self._initialize_interface()
 
-    super(TaskQueue, self).__init__(n_threads) # creates self._queue
-
-  @property
-  def queue_name(self):
-    return self._queue_name
-  
-  @property
-  def queue_server(self):
-    return self._queue_server
-
   # This is key to making sure threading works. Don't refactor this method away.
   def _initialize_interface(self):
     server = self._queue_server.lower()
@@ -84,11 +63,18 @@ class TaskQueue(ThreadedQueue):
     else:
       raise NotImplementedError('Unknown server ' + self._queue_server)
 
+  @property
+  def queue_name(self):
+    return self._queue_name
+  
+  @property
+  def queue_server(self):
+    return self._queue_server
+
   @property
   def enqueued(self):
     """
     Returns the approximate(!) number of tasks enqueued in the cloud.
-
     WARNING: The number computed by Google is eventually
       consistent. It may return impossible numbers that
       are small deviations from the number in the queue.
@@ -98,27 +84,6 @@ class TaskQueue(ThreadedQueue):
     Returns: (int) number of tasks in cloud queue
     """
     return self._api.enqueued
-    
-  def insert(self, task, args=[], kwargs={}):
-    """
-    Insert a task into an existing queue.
-    """
-    body = {
-      "payload": task.payload(),
-      "queueName": self._queue_name,
-      "groupByTag": True,
-      "tag": task.__class__.__name__
-    }
-
-    def cloud_insertion(api):
-      api.insert(body)
-
-    if len(self._threads):
-      self.put(cloud_insertion)
-    else:
-      cloud_insertion(self._api)
-
-    return self
 
   def status(self):
     """
@@ -162,7 +127,7 @@ class TaskQueue(ThreadedQueue):
     )
 
     if not len(tasks):
-      raise TaskQueue.QueueEmpty
+      raise QueueEmpty
 
     task = tasks[0]
     return totask(task)
@@ -189,35 +154,6 @@ class TaskQueue(ThreadedQueue):
         self.wait()
       return self
 
-  def acknowledge(self, task_id):
-    if isinstance(task_id, RegisteredTask):
-      task_id = task_id.id
-
-    def cloud_delete(api):
-      api.acknowledge(task_id)
-
-    if len(self._threads):
-      self.put(cloud_delete)
-    else:
-      cloud_delete(self._api)
-
-    return self
-
-  def delete(self, task_id):
-    """Deletes a task from a TaskQueue."""
-    if isinstance(task_id, RegisteredTask):
-      task_id = task_id.id
-
-    def cloud_delete(api):
-      api.delete(task_id)
-
-    if len(self._threads):
-      self.put(cloud_delete)
-    else:
-      cloud_delete(self._api)
-
-    return self
-
   def poll(
     self, lease_seconds=LEASE_SECONDS, tag=None, 
     verbose=False, execute_args=[], execute_kwargs={}, 
@@ -228,7 +164,6 @@ class TaskQueue(ThreadedQueue):
     Poll a queue until a stop condition is reached (default forever). Note
     that this function is not thread safe as it requires a global variable
     to intercept SIGINT.
-
     lease_seconds: each task should be leased for this many seconds
     tag: if specified, query for only tasks that match this tag
     execute_args / execute_kwargs: pass these arguments to task execution
@@ -244,7 +179,6 @@ class TaskQueue(ThreadedQueue):
       once after every task.
     log_fn: Feed error messages to this function, default print (when verbose is enabled).
     verbose: print out the status of each step
-
     Return: number of tasks executed
     """
     global LOOP
@@ -283,7 +217,7 @@ class TaskQueue(ThreadedQueue):
     executed = 0
 
     backoff = False
-    backoff_exceptions = tuple(list(backoff_exceptions) + [ TaskQueue.QueueEmpty ])
+    backoff_exceptions = tuple(list(backoff_exceptions) + [ QueueEmpty ])
 
     while LOOP:
       task = 'unknown' # for error message prior to leasing
@@ -320,6 +254,284 @@ class TaskQueue(ThreadedQueue):
     while self.enqueued > 0:
       time.sleep(interval_sec)
 
+class TaskQueue(SuperTaskQueue, ThreadedQueue):
+  """
+  The standard usage is that a client calls lease to get the next available task,
+  performs that task, and then calls task.delete on that task before the lease expires.
+  If the client cannot finish the task before the lease expires,
+  and has a reasonable chance of completing the task,
+  it should call task.update before the lease expires.
+  If the client completes the task after the lease has expired,
+  it still needs to delete the task. 
+  Tasks should be designed to be idempotent to avoid errors 
+  if multiple clients complete the same task.
+  """
+  def __init__(
+    self, queue_name=QUEUE_NAME, queue_server=QUEUE_TYPE, 
+    region=None, qurl=None, n_threads=40, project=PROJECT_NAME
+  ):
+
+    SuperTaskQueue.__init__(
+      self, queue_name, queue_server, 
+      region, qurl, n_threads, project
+    )
+    ThreadedQueue.__init__(self, n_threads) # creates self._queue
+    
+  def insert(self, task, args=[], kwargs={}, delay_seconds=0):
+    """
+    Insert a task into an existing queue.
+    """
+    body = {
+      "payload": task.payload(),
+      "queueName": self._queue_name,
+      "groupByTag": True,
+      "tag": task.__class__.__name__
+    }
+
+    def cloud_insertion(api):
+      api.insert(body, delay_seconds)
+
+    if len(self._threads):
+      self.put(cloud_insertion)
+    else:
+      cloud_insertion(self._api)
+
+    return self
+
+  def insert_all(self, tasks, delay_seconds=0, total=None):
+    if total is None:
+      try:
+        total = len(tasks)
+      except TypeError:
+        pass
+     
+    batch_size = 10 
+    def genbatches(itr):
+      while True:
+        batch = []
+        try:
+          for i in range(batch_size):
+            batch.append(next(itr))
+        except StopIteration:
+          pass
+
+        if len(batch) == 0:
+          raise StopIteration
+
+        yield batch
+
+    bodies = (
+      {
+        "payload": task.payload(),
+        "queueName": self._queue_name,
+        "groupByTag": True,
+        "tag": task.__class__.__name__
+      } 
+
+      for task in tasks
+    )
+
+    def cloud_insertion(batch):
+      self._api.insert(batch, delay_seconds)
+
+    fns = ( partial(cloud_insertion, batch) for batch in genbatches(bodies) )
+
+    schedule_threaded_jobs(
+      fns=fns,
+      concurrency=20,
+      progress='Inserting',
+      total=total,
+      batch_size=batch_size,
+    )
+
+  def acknowledge(self, task_id):
+    if isinstance(task_id, RegisteredTask):
+      task_id = task_id.id
+
+    def cloud_delete(api):
+      api.acknowledge(task_id)
+
+    if len(self._threads):
+      self.put(cloud_delete)
+    else:
+      cloud_delete(self._api)
+
+    return self
+
+  def delete(self, task_id):
+    """Deletes a task from a TaskQueue."""
+    if isinstance(task_id, RegisteredTask):
+      task_id = task_id.id
+
+    def cloud_delete(api):
+      api.delete(task_id)
+
+    if len(self._threads):
+      self.put(cloud_delete)
+    else:
+      cloud_delete(self._api)
+
+    return self
+
+class GreenTaskQueue(SuperTaskQueue):
+  """
+  The standard usage is that a client calls lease to get the next available task,
+  performs that task, and then calls task.delete on that task before the lease expires.
+  If the client cannot finish the task before the lease expires,
+  and has a reasonable chance of completing the task,
+  it should call task.update before the lease expires.
+
+  If the client completes the task after the lease has expired,
+  it still needs to delete the task. 
+
+  Tasks should be designed to be idempotent to avoid errors 
+  if multiple clients complete the same task.
+  """
+  def __init__(
+    self, queue_name=QUEUE_NAME, queue_server=QUEUE_TYPE, 
+    region=None, qurl=None, n_threads=40, project=PROJECT_NAME
+  ):
+    n_threads = max(n_threads, 1)
+    SuperTaskQueue.__init__(
+      self, queue_name, queue_server, 
+      region, qurl, n_threads, project
+    )
+    self._pool = gevent.pool.Pool(n_threads)
+    self.check_monkey_patch_status()
+
+
+  def check_monkey_patch_status(self):
+    import gevent.monkey
+    if not gevent.monkey.is_module_patched("socket"):
+      print(yellow("""
+    Switching ThreadedQueue to using green threads. This requires
+    monkey patching the standard library to use a cooperative 
+    non-blocking threading model.
+
+    Please place the following lines at the beginning of your
+    program:
+
+    import gevent.monkey
+    gevent.monkey.patch_all()
+        """))
+
+  def insert(self, task, args=[], kwargs={}, delay_seconds=0):
+    """
+    Insert a task into an existing queue.
+    """
+    body = {
+      "payload": task.payload(),
+      "queueName": self._queue_name,
+      "groupByTag": True,
+      "tag": task.__class__.__name__
+    }
+
+    def cloud_insertion():
+      self._api.insert(body, delay_seconds)
+
+    self._pool.spawn(cloud_insertion)
+
+    return self
+
+  def insert_all(self, tasks, delay_seconds=0, total=None):
+    if total is None:
+      try:
+        total = len(tasks)
+      except TypeError:
+        pass
+     
+    batch_size = 10 
+    def genbatches(itr):
+      while True:
+        batch = []
+        try:
+          for i in range(batch_size):
+            batch.append(next(itr))
+        except StopIteration:
+          pass
+
+        if len(batch) == 0:
+          raise StopIteration
+
+        yield batch
+
+    bodies = (
+      {
+        "payload": task.payload(),
+        "queueName": self._queue_name,
+        "groupByTag": True,
+        "tag": task.__class__.__name__
+      } 
+
+      for task in tasks
+    )
+
+    def cloud_insertion(batch):
+      self._api.insert(batch, delay_seconds)
+
+    fns = ( partial(cloud_insertion, batch) for batch in genbatches(bodies) )
+
+    schedule_green_jobs(
+      fns=fns,
+      concurrency=20,
+      progress='Inserting',
+      total=total,
+      batch_size=batch_size,
+    )
+
+  def status(self):
+    """
+    Gets information about the TaskQueue
+    """
+    return self._api.get(getStats=True)
+
+  def get_task(self, tid):
+    """
+    Gets the named task in the TaskQueue. 
+    tid is a unique string Google provides 
+    e.g. '7c6e81c9b7ab23f0'
+    """
+    return self._api.get(tid)
+
+  def list(self):
+    """
+    Lists all non-deleted Tasks in a TaskQueue, 
+    whether or not they are currently leased, up to a maximum of 100.
+    """
+    return [ totask(x) for x in self._api.list() ]
+
+  def acknowledge(self, task_id):
+    if isinstance(task_id, RegisteredTask):
+      task_id = task_id.id
+
+    def cloud_delete():
+      self._api.acknowledge(task_id)
+
+    self._pool.spawn(cloud_delete)
+
+    return self
+
+  def delete(self, task_id):
+    """Deletes a task from a TaskQueue."""
+    if isinstance(task_id, RegisteredTask):
+      task_id = task_id.id
+
+    def cloud_delete():
+      self._api.delete(task_id)
+
+    self._pool.spawn(cloud_delete)
+
+    return self
+
+  def wait(self):
+    self._pool.join()
+
+  def __enter__(self):
+    return self
+
+  def __exit__(self, exception_type, exception_value, traceback):
+    self._pool.join()
+
 class MockTaskQueue(object):
   def __init__(self, *args, **kwargs):
     pass
@@ -333,6 +545,10 @@ class MockTaskQueue(object):
     task.execute(*args, **kwargs)
     del task
 
+  def insert_all(self, tasks, delay_seconds=0, total=None):
+    for task in tasks:
+      self.insert(task)
+
   def poll(self, *args, **kwargs):
     return self
 
@@ -366,7 +582,12 @@ class LocalTaskQueue(object):
 
     self.queue.append( (task, args, kwargs) )
 
+  def insert_all(self, tasks, delay_seconds=0, total=None):
+    for task in tasks:
+      self.queue.append(tasks)
+
   def wait(self, progress=None):
+    self._process(progress)
     return self
 
   def poll(self, *args, **kwargs):
@@ -375,15 +596,24 @@ class LocalTaskQueue(object):
   def kill_threads(self):
     return self
 
+  def _process(self, progress=True):
+    with tqdm(total=len(self.queue), desc="Tasks", disable=(not progress)) as pbar:
+      if self.parallel == 1:
+        while self.queue:
+          _task_execute(self.queue.pop(0))
+          pbar.update()
+      else:
+        with concurrent.futures.ProcessPoolExecutor(max_workers=self.parallel) as executor:
+          for _ in executor.map(_task_execute, self.queue):
+            pbar.update()
+    
+      self.queue = []
+
   def __enter__(self):
     return self
 
   def __exit__(self, exception_type, exception_value, traceback):
-    with tqdm(total=len(self.queue), desc="Tasks") as pbar:
-      with concurrent.futures.ProcessPoolExecutor(max_workers=self.parallel) as executor:
-        for _ in executor.map(_task_execute, self.queue):
-          pbar.update()
-    self.queue = []
+    self._process()
 
 # Necessary to define here to make the 
 # function picklable
