@@ -6,7 +6,7 @@ from six.moves import range
 import pytest
 
 import taskqueue
-from taskqueue import RegisteredTask, TaskQueue, MockTask, PrintTask, LocalTaskQueue, MockTaskQueue
+from taskqueue import RegisteredTask, GreenTaskQueue, TaskQueue, MockTask, PrintTask, LocalTaskQueue, MockTaskQueue
 from taskqueue import QUEUE_NAME
 
 TRAVIS_BRANCH = None if 'TRAVIS_BRANCH' not in os.environ else os.environ['TRAVIS_BRANCH']
@@ -54,70 +54,52 @@ def test_get():
   global QUEUE_NAME
 
   for qtype in QTYPES:
-    tq = TaskQueue(n_threads=0, queue_name=QUEUE_NAME, queue_server=qtype, qurl=QURL)
+    for QueueClass in (TaskQueue, GreenTaskQueue):
+      tq = QueueClass(n_threads=0, queue_name=QUEUE_NAME, queue_server=qtype, qurl=QURL)
 
-    n_inserts = 5
-    tq.purge()
-    for _ in range(n_inserts):
-      task = PrintTask()
-      tq.insert(task)
-    tq.wait()
-    tq.purge()
+      n_inserts = 5
+      tq.purge()
+      for _ in range(n_inserts):
+        task = PrintTask()
+        tq.insert(task)
+      tq.wait()
+      tq.purge()
 
 def test_single_threaded_insertion():
   global QUEUE_NAME
 
   for qtype in QTYPES:
-    tq = TaskQueue(n_threads=0, queue_name=QUEUE_NAME, queue_server=qtype, qurl=QURL)
+    for QueueClass in (TaskQueue, GreenTaskQueue):
+      tq = QueueClass(n_threads=0, queue_name=QUEUE_NAME, queue_server=qtype, qurl=QURL)
 
-    tq.purge()
+      tq.purge()
+      
+      n_inserts = 5
+      for _ in range(n_inserts):
+        task = PrintTask()
+        tq.insert(task)
+      tq.wait()
 
-    if qtype != 'aws':
-      assert tq.enqueued == 0
-    
-    n_inserts = 5
-    for _ in range(n_inserts):
-      task = PrintTask()
-      tq.insert(task)
-    tq.wait()
+      assert all(map(lambda x: type(x) == PrintTask, tq.list()))
 
-    if qtype != 'aws':
-      assert tq.enqueued == n_inserts
-
-    assert all(map(lambda x: type(x) == PrintTask, tq.list()))
-
-    tq.purge()
-    
-    if qtype != 'aws':
-      assert tq.enqueued == 0
+      tq.purge()
 
 def test_multi_threaded_insertion():
   global QUEUE_NAME
   for qtype in QTYPES:
-    tq = TaskQueue(n_threads=40, queue_name=QUEUE_NAME, queue_server=qtype, qurl=QURL)
+    for QueueClass in (TaskQueue, GreenTaskQueue):
+      tq = QueueClass(n_threads=0, queue_name=QUEUE_NAME, queue_server=qtype, qurl=QURL)
 
-    n_inserts = 100
-    tq.purge()
-    tq.wait()
-    if qtype != 'aws':
-      assert tq.enqueued == 0
-    
-    for _ in range(n_inserts):
-      task = PrintTask()
-      tq.insert(task)
-    tq.wait()
+      n_inserts = 10
+      tq.purge()
+      tq.wait()
+     
+      for _ in range(n_inserts):
+        task = PrintTask()
+        tq.insert(task)
+      tq.wait()
 
-    if qtype == 'aws':
-      list_len = 10
-    else:
-      list_len = 100
-
-    lst = tq.list()
-    assert len(lst) == list_len # task list api only lists 100 items at a time
-    assert all(map(lambda x: type(x) == PrintTask, lst))
-    tq.purge()
-    if qtype != 'aws':
-      assert tq.enqueued == 0
+      tq.purge()
 
 # def test_multiprocess_upload():
 #   global QURL
@@ -142,8 +124,9 @@ def test_multi_threaded_insertion():
 def test_400_errors():
   global QUEUE_NAME
   for qtype in QTYPES:
-    with TaskQueue(n_threads=1, queue_name=QUEUE_NAME, queue_server=qtype, qurl=QURL) as tq:
-      tq.delete('nonexistent')
+    for QueueClass in (TaskQueue, GreenTaskQueue):
+      with QueueClass(n_threads=0, queue_name=QUEUE_NAME, queue_server=qtype, qurl=QURL) as tq:
+        tq.delete('nonexistent')
 
 def test_local_taskqueue():
   with LocalTaskQueue(parallel=True, progress=False) as tq:
@@ -152,6 +135,10 @@ def test_local_taskqueue():
         MockTask(arg=i)
       )
 
+  with LocalTaskQueue(parallel=1, progress=False) as tq:
+    for i in range(200):
+      tq.insert(ExecutePrintTask(), [i], { 'wow2': 4 })
+
   with LocalTaskQueue(parallel=True, progress=False) as tq:
     for i in range(200):
       tq.insert(ExecutePrintTask(), [i], { 'wow2': 4 })
